/*  Copyright (C) 2008 Joshua Judson Rosen <rozzin@geekspace.com>.

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 3 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING. If not, see
    <http://www.gnu.org/licenses/> or write to:

        The Free Software Foundation, inc.
        51 Franklin Street, Fifth Floor
        Boston, MA 02110-1301
        USA
*/

#include <stdlib.h>
#include "constants.h"
#include "random.h"

#include <math.h>
#include <cairo.h>
#include "graphics.h"

#include "radial.h"
#include "spiral.h"
#include "line.h"
#include "figure.h"
#include "metashape.h"
#include "shape.h"
#include "path.h"
#include "symmetry.h"

static generator_fn *terminals[] = { gen_line /*, gen_scribble */ };
static const int nterminals = (sizeof terminals)/(sizeof (generator_fn *));

       generator_fn *children[] = { gen_radial,
                                    gen_figure,
                                    gen_shape,
                                    gen_path,
                                    gen_line,
                                    gen_symmetry,
                                    gen_spiral,
       };
       const int nchildren = (sizeof children)/(sizeof (generator_fn *));


static struct generator *
rnd_pick_terminal (int level)
{
  return terminals[(int) floor (rndin (0, (double) nterminals))] (level);
}

static struct generator *
rnd_pick_child (int level)
{
  return children[(int) floor (rndin (0, (double) nchildren))] (level);
}

struct generator *
generate_child (int level)
{
  if (level < recursion_max) {
    return rnd_pick_child (level);
  } else {
    return rnd_pick_terminal (level);
  }
}

void
exec_generator (struct generator *generator, cairo_t *cr)
{
  cairo_save (cr);
  generator->produce (generator, cr);
  cairo_restore (cr);
}

void
visualid_draw (struct generator *gen, cairo_t *cr,
               double width, double height,
               double linewidth)
{
  /* The `virtual canvas' has a unit radius,
     which means it's 2 units wide and 2 units tall: */
  const double target_size = 2;

  double path_left, path_right,
         path_top, path_bottom;
  double stroke_left, stroke_right,
         stroke_top, stroke_bottom;
  double margin_left, margin_right,
         margin_top, margin_bottom;
  double path_width, path_height;
  double midpoint_x, midpoint_y;
  double rescale;

  cairo_save (cr);

  cairo_scale (cr, width/target_size, -height/target_size);
  cairo_translate (cr, target_size/2, -target_size/2);

  exec_generator (gen, cr);

  cairo_path_extents (cr, &path_left, &path_bottom,
                          &path_right, &path_top);
  cairo_set_line_width (cr, linewidth);
  cairo_stroke_extents (cr, &stroke_left, &stroke_bottom,
                            &stroke_right, &stroke_top);

  /* These need to be computed individually by comparing stroke-bounds
     and path-bounds, because the space needed for non-round
     line-joints is unpredictable: */

  margin_top = stroke_top - path_top;
  margin_left = path_left - stroke_left;
  margin_right = stroke_right - path_right;
  margin_bottom = path_bottom - stroke_bottom;

  path_width = path_right - path_left;
  path_height = path_top - path_bottom;

  /* Scale the glyph up or down so that it fits perfectly onto the canvas,
     (note that stroking the path with a brush that's (linewidth)
     units wide will add (linewidth/2) to each side,
     so the target-size of the _path_ per se should be that much smaller),
     assuming a circular brush: */
  rescale = (target_size - fmax ((margin_left + margin_right),
                                 (margin_bottom + margin_top)))
    / fmax (path_width, path_height);

  stroke_top = path_top + margin_top/rescale;
  stroke_left = path_left - margin_left/rescale;
  stroke_right = path_right + margin_right/rescale;
  stroke_bottom = path_bottom - margin_bottom/rescale;

  cairo_new_path (cr);

  cairo_scale (cr, rescale, rescale);

  midpoint_x = (stroke_left + stroke_right)/2;
  midpoint_y = (stroke_bottom + stroke_top)/2;
  cairo_translate (cr, -midpoint_x, -midpoint_y);

  exec_generator (gen, cr);

  cairo_restore (cr);
}

/*
  Want to mutate a base generator based on:
  - how different the new string is from the base string

  What does the `different' mean?

  The `sameness' is characterised by the length of the string in common.

  string1_len / (string1_start + 1)

  or: 1 - score

  o: ...
 */
/* struct generator * */
/* mutate_generator (struct generator *base, int level, double amount) */
/* { */
/*   mutator_fn *mutate = get_mutator (base, level, amount); */
/*   struct generator *new_gen = mutate (base, level, amount); */
/*   return new_gen; */
/* } */
