/*  Copyright (C) 2008 Joshua Judson Rosen <rozzin@geekspace.com>.

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 3 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING. If not, see
    <http://www.gnu.org/licenses/> or write to:

        The Free Software Foundation, inc.
        51 Franklin Street, Fifth Floor
        Boston, MA 02110-1301
        USA
*/

#include <stdlib.h>
#include "constants.h"
#include "random.h"

#include <math.h>
#include <cairo.h>

#include <glib.h>
#include "graphics.h"
#include "metashape.h"

static double
outer_angle (double r, double r_plus, double angle_inside)
{
      double y_1 = r_plus * cos (angle_inside) - r;
      double x = r_plus * sin (angle_inside);
      double theta_1 = atan (x / y_1);

      if (theta_1 < 0) {
        theta_1 += M_PI;
      }

      return theta_1;
}

static void
exec_metashape (struct shape_generator *shape, cairo_t *cr)
{
  double x, y;
  int i;

  cairo_get_current_point (cr, &x, &y);
  cairo_new_sub_path (cr);

  cairo_save (cr);
  for (i = 0; i < shape->n; i++) {
    cairo_line_to (cr, 0, shape->points[i]);
    if (shape->outline_child) {
      cairo_save (cr);
      cairo_translate (cr, 0, shape->points[i]);
      cairo_scale (cr, shape->oc_scale, shape->oc_scale);
      shape->outline_child->produce (shape->outline_child, cr);
      cairo_restore (cr);
    }

    if (shape->inside_child) {
      double r = shape->points[i];
      double r_plus = shape->points[(i + 1) % shape->n];
      double r_minus = shape->points[(i - 1) % shape->n];
      double angle_inside = shape->max_angle / shape->n;

      cairo_save (cr);
      cairo_translate (cr, 0, shape->points[i]);
      cairo_rotate (cr, (outer_angle (r, r_plus, angle_inside) -
                         outer_angle (r, r_minus, angle_inside)) / 2);

      cairo_translate (cr, 0, -shape->ic_scale*4);

      cairo_scale (cr, shape->ic_scale, shape->ic_scale);
      shape->inside_child->produce (shape->inside_child, cr);
      cairo_restore (cr);
    }

    cairo_rotate (cr, shape->max_angle / shape->n);
  }
  cairo_restore (cr);

  if (shape->max_angle == 2*M_PI) {
    cairo_line_to (cr, 0, shape->points[0]);
    cairo_close_path (cr);
  }

  cairo_new_sub_path (cr);
  cairo_move_to (cr, x, y);
}

struct shape_generator *
gen_metashape (int level, double arclen)
{
  struct shape_generator *g = g_new (struct shape_generator, 1);

  int i, j;

  ((struct generator *) g)->produce = (producer_fn *) exec_metashape;

  g->n = pow (2, rndin_level (3, 6, level));

  g->fourier = rndprob (0.5); /* Whether to use fourier-series generation */

  /* Parameters for fourier-series generation: */
  g->ncoef = 2 * rndin (2, 5) + 1;
  g->coefficients = g_new (double, g->ncoef);

  /* spectral exponent for FFT filtering: */
/*   g->exponent = rndin (1.6, 2.2); */

  g->points = g_new (double, g->n);

  g->max_angle = arclen;

  /* XXX To be true to the essay, we should check whether g->fourier,
     rather than just always using fourier-series generation: */

  for (i = 0; i < g->n; i++) {
    g->points[i] = .5; /* run around on a .5-unit circle;
                          starting with 1 will probably result in
                          excessively large shapes. */
  }

  for (j = 0; j < g->ncoef; j++) {
    g->coefficients[j] = rnd11 () / (j + 1); /* 1/f envelope */

    for (i = 0; i < g->n; i++) {
      double angle = g->max_angle * i / g->n;

      g->points[i] += g->coefficients[j] * sin((j + 1) * angle) / 2;
    }
  }
  /* Alternately, we don't really /need/ to use a fourier
     series--whatever series we generate will be cyclic by virtue of
     the ends being connected. This produces fine shapes, for example:

         g->points[i] = 1 - rnd11 () / ((double) i + 1);

     ... though it probably ends up ~100% too big
     (but sometimes this looks good).
  */


  if (g->max_angle < 2*M_PI
  ||  rndprob_level (.6, level)) {
    g->outline_child = generate_child (1 + level);
    g->oc_scale = rndin (.1, .6);
  } else {
    g->outline_child = NULL;
  }
  if (g->max_angle < 2*M_PI
  ||  !rndprob_level (.6, level)) {
    g->inside_child = NULL;
  } else {
    g->inside_child = generate_child (1 + level);
    g->ic_scale = rndin (.1, 2) / g->n;
  }

  return g;
}
